#include <io.h>
#include <system.h>
#include <stdio.h>
#include <stdlib.h>

#include "wiimote.h"
#include "wiimote_api.h"

// vga settings in hardware
#define HPAD 96+48
#define VPAD 2+33
#define EXTRA_LEFT 800
#define EXTRA_TOP 525
#define SCREEN_LEFT EXTRA_LEFT+HPAD
#define SCREEN_TOP EXTRA_TOP+VPAD
#define SCREEN_CENTER_X SCREEN_LEFT+320
#define SCREEN_CENTER_Y SCREEN_TOP+240

// game constants
#define TILE_SIZE 32
#define MAX_MAZE_SIZE_X 16  // in # of tiles
#define MAX_MAZE_SIZE_Y 16  // in # of tiles

// settings
#define PAUSE_AMT 40000000     // in # of for loops
#define SPEED_PIX 2         // should be 1, 2, 4, or 8

// memory loc offsets
#define MAZE_X_OFFSET 0             // in pixels
#define MAZE_Y_OFFSET 1             // in pixels
#define CHAR_X_OFFSET 2             // in pixels
#define CHAR_Y_OFFSET 3             // in pixels
#define GOAL_X_OFFSET 4             // in maze tiles
#define GOAL_Y_OFFSET 5             // in maze tiles
#define WINDOW_SIZE_OFFSET 6        // in maze tiles
#define WHICH_KEYS_OFFSET 7         // see key defs
#define MAZE_TILES_OFFSET 8         // see maze tile types
#define OBJ_TILES_OFFSET 73         // see obj tile types
#define WHICH_CHAR_TILE_OFFSET 137  // which char tile to draw



// due to poor foresight, it will be very annoying to add more keys now
// so just leave these as is unless we really want more keys 
#define KEY1 1
#define KEY2 2
#define KEY3 4

// maze tile types
#define LEFT_WALL 0
#define TOP_WALL 1
#define RIGHT_WALL 2
#define BOTTOM_WALL 3
#define TOP_LEFT_CORNER 4
#define TOP_RIGHT_CORNER 5
#define BOTTOM_RIGHT_CORNER 6
#define BOTTOM_LEFT_CORNER  7
#define VERTICAL_HALL 8
#define HORIZONTAL_HALL 9
#define BOTTOM_DEADEND 10
#define LEFT_DEADEND 11
#define TOP_DEADEND 12
#define RIGHT_DEADEND 13
#define NO_MAZE_TILE 14

// obj tile types
#define LEFT_WALL_RED 0
#define TOP_WALL_RED 1
#define RIGHT_WALL_RED 2
#define BOTTOM_WALL_RED 3
#define LEFT_WALL_GREEN 4
#define TOP_WALL_GREEN 5
#define RIGHT_WALL_GREEN 6
#define BOTTOM_WALL_GREEN 7
#define LEFT_WALL_BLUE 8
#define TOP_WALL_BLUE 9
#define RIGHT_WALL_BLUE 10
#define BOTTOM_WALL_BLUE 11
#define KEY_RED 12
#define KEY_GREEN 13
#define KEY_BLUE 14
#define NO_OBJ_TILE 15

// level descriptor offsets
#define DESC_BASIC_OFFSET 0
#define DESC_MAZE_OFFSET 1
#define DESC_OBJ_OFFSET MAX_MAZE_SIZE_Y+1

// basic level descriptor offsets
#define DESC_BASIC_XPOS_MAZE_OFFSET 0
#define DESC_BASIC_YPOS_MAZE_OFFSET 1
#define DESC_BASIC_XPOS_CHAR_OFFSET 2
#define DESC_BASIC_YPOS_CHAR_OFFSET 3
#define DESC_BASIC_XPOS_GOAL_OFFSET 4
#define DESC_BASIC_YPOS_GOAL_OFFSET 5
#define DESC_BASIC_HALF_XWIN_OFFSET 6
#define DESC_BASIC_HALF_YWIN_OFFSET 7

// character movements directions
#define UP 0
#define DOWN 1
#define LEFT 2
#define RIGHT 3

// Number of levels
#define NUM_LEVELS 4

// function declarations
void move_maze_to(int x, int y);
int get_maze_x();
int get_maze_y();

void move_char_to_pix(int x, int y);
void move_char_to_cell(int col, int row);
int get_char_start_x(int **the_level);
int get_char_start_y(int **the_level);

void set_goal_at(int col, int row);

void hide_keys_and_doors(int which);
void show_keys_and_doors(int which);

void set_window_size(int half_cols, int half_rows);

void animate(int dir, wiimote_t *wiimote);
void toggle_char_tile();

void get_level(int **the_level, int level);
void get_level1(int **the_level);
void get_level2(int **the_level);
void get_level3(int **the_level);
void get_level4(int **the_level);
// add more get_levelX fns here and have them be called by get_level

void create_level(int **level_desc);

void set4maze_tiles(int offset, int tile1, int tile2, int tile3, int tile4);
void set4obj_tiles(int offset, int tile1, int tile2, int tile3, int tile4);

int can_move(int **the_level, int fromx, int fromy, int dir);
int contains_wall(int tile_type, int dir);
int contains_door(int tile_type, int dir);

int is_at_goal_loc(int **the_level, int x, int y);
int ontop_of_which_key(int **the_level, int x, int y);

int get_next_move(wiimote_t *wiimote);

// wiimote sound fns
void make_got_to_goal_sound(wiimote_t *wiimote);
void make_got_key_sound(wiimote_t *wiimote);
void make_moving_sound();
void make_hit_wall_sound(wiimote_t *wiimote);

void pause();
void animate_pause(wiimote_t *wiimote);


/* sounds */
uint8_t sample[20] = {
	0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,
	0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c,0x3c
};

int main(int argc, char **argv)
{               
    int i;
    int count=0;

    wiimote_t wiimote = WIIMOTE_INIT;

    /* check command line args */
    if(argc<2){
      printf("You didn't provide a remote address.\nUse \'hcitool sca\'n to get the address of your wiimote\n");
      exit(1);
    }
    
    /* connect the wiimote */
    if(wiimote_connect(&wiimote, argv[1])<0){
      printf("Unable to connect to wiimote: %s\n",wiimote_get_error());
      exit(1);
    }
    
    printf("Connection established!\n");
    wiimote.led.one = 1;
    wiimote_speaker_init(&wiimote, WIIMOTE_FMT_S4, 0xff);
    
    // set up level descriptor               
    int **my_level = NULL;
    my_level = (int **)malloc((MAX_MAZE_SIZE_Y * 2 + 1)*sizeof(int *));
    my_level[DESC_BASIC_OFFSET] = (int *)malloc(8*sizeof(int));
    for( i = DESC_MAZE_OFFSET; i < DESC_MAZE_OFFSET+MAX_MAZE_SIZE_Y; i++ )
        my_level[i] = (int *)malloc(MAX_MAZE_SIZE_X * sizeof(int));
    for( i = DESC_OBJ_OFFSET; i < DESC_OBJ_OFFSET+MAX_MAZE_SIZE_Y; i++ )
        my_level[i] = (int *)malloc(MAX_MAZE_SIZE_X * sizeof(int));
    
    // game logic
    int char_x, char_y, move_dir, which_key, cur_level = 1;


    while(1) {
        get_level(my_level,cur_level);
        create_level(my_level);
        
        show_keys_and_doors(KEY1);
        show_keys_and_doors(KEY2);
        show_keys_and_doors(KEY3);
        
        char_x = get_char_start_x(my_level);
        char_y = get_char_start_y(my_level);
    
        while(1) {
	  fprintf(stderr, "iteration: %d:\n", count);
	  count++;
	  move_dir = get_next_move(&wiimote);
	  if( can_move(my_level,char_x,char_y,move_dir) ) {
	    // Used to animate. Helps clear input queue of incoming packets via wiimote_update
	    animate(move_dir, &wiimote); 
	                
	    if( move_dir == UP ) char_y--;
	    else if( move_dir == DOWN ) char_y++;
	    else if( move_dir == LEFT ) char_x--;
	    else if( move_dir == RIGHT ) char_x++;
            
	    which_key = ontop_of_which_key(my_level,char_x,char_y);
	    if( which_key != -1 ) {
	      hide_keys_and_doors(which_key);
	      make_got_key_sound(&wiimote);
	    }
            
	    if( is_at_goal_loc( my_level, char_x, char_y ) ) {
	      make_got_to_goal_sound(&wiimote);
	      cur_level = cur_level+1;
	      if(cur_level > NUM_LEVELS)
		cur_level = 1;
	      break;
	    }
	  }
	  else
	    make_hit_wall_sound(&wiimote);
        }
    }
    
}

void move_maze_to(int x, int y) {
  IOWR(VGA_BASE,MAZE_X_OFFSET,x);
  IOWR(VGA_BASE,MAZE_Y_OFFSET,y);
}

int get_maze_x() {
  return IORD(VGA_BASE,MAZE_X_OFFSET);
}

int get_maze_y() {
  return IORD(VGA_BASE,MAZE_Y_OFFSET);
}

void move_char_to_pix(int x, int y) {
    IOWR(VGA_BASE,CHAR_X_OFFSET,x);
    IOWR(VGA_BASE,CHAR_Y_OFFSET,y);
}

// be sure to set maze where you want it first
void move_char_to_cell(int col, int row) {
    IOWR(VGA_BASE,CHAR_X_OFFSET,get_maze_x()+col*TILE_SIZE);
    IOWR(VGA_BASE,CHAR_Y_OFFSET,get_maze_y()+row*TILE_SIZE);
}

void set_goal_at(int col, int row) {
    IOWR(VGA_BASE,GOAL_X_OFFSET,col);
    IOWR(VGA_BASE,GOAL_Y_OFFSET,row);
}

void hide_keys_and_doors(int which) {
    IOWR(VGA_BASE,WHICH_KEYS_OFFSET,IORD(VGA_BASE,WHICH_KEYS_OFFSET)&(~which));
}

void show_keys_and_doors(int which) {
    IOWR(VGA_BASE,WHICH_KEYS_OFFSET,IORD(VGA_BASE,WHICH_KEYS_OFFSET)|which);
}

int has_key(int which) {
    return !(IORD(VGA_BASE,WHICH_KEYS_OFFSET)&which);
}

// pass to this function HALF of the size of the window you want in number of tiles
// i.e. if you want a 4x6 window, pass 2 and 3 
// these should be at most 8 bit numbers
void set_window_size(int half_cols, int half_rows) {
    IOWR(VGA_BASE,WINDOW_SIZE_OFFSET,(half_cols<<8)|half_rows);
}

void set4maze_tiles(int offset, int tile1, int tile2, int tile3, int tile4) {
    IOWR(VGA_BASE,MAZE_TILES_OFFSET+offset,(tile1<<12)|(tile2<<8)|(tile3<<4)|(tile4));
}

void set4obj_tiles(int offset, int tile1, int tile2, int tile3, int tile4) {
    IOWR(VGA_BASE,OBJ_TILES_OFFSET+offset,(tile1<<12)|(tile2<<8)|(tile3<<4)|(tile4));
}

void create_level(int **level_desc) {
    // set basic info
    set_window_size(level_desc[DESC_BASIC_OFFSET][DESC_BASIC_HALF_XWIN_OFFSET],level_desc[DESC_BASIC_OFFSET][DESC_BASIC_HALF_YWIN_OFFSET]);
    move_maze_to(level_desc[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_MAZE_OFFSET],level_desc[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_MAZE_OFFSET]); pause(); pause(); pause(); pause(); // pauses are stupid hacks
    move_char_to_cell(level_desc[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_CHAR_OFFSET],level_desc[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_CHAR_OFFSET]);
    set_goal_at(level_desc[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_GOAL_OFFSET],level_desc[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_GOAL_OFFSET]); pause();
    
    // set maze and obj tiles
    int i,j, off;
    for(i = 0, off = 0; i < MAX_MAZE_SIZE_Y; i++ ) {
        for( j = 0; j < MAX_MAZE_SIZE_X; j+=4, off++ ) {
	  set4maze_tiles(off, level_desc[DESC_MAZE_OFFSET+i][j],level_desc[DESC_MAZE_OFFSET+i][j+1],level_desc[DESC_MAZE_OFFSET+i][j+2],level_desc[DESC_MAZE_OFFSET+i][j+3]);
            set4obj_tiles(off, level_desc[DESC_OBJ_OFFSET+i][j],level_desc[DESC_OBJ_OFFSET+i][j+1],level_desc[DESC_OBJ_OFFSET+i][j+2],level_desc[DESC_OBJ_OFFSET+i][j+3]);
        }
    }
}

void get_level(int **the_level, int level)
{
    if( level == 1 ) get_level1(the_level);
    else if( level == 2 ) get_level2(the_level);
    else if( level == 3 ) get_level3(the_level);
    else if( level == 4 ) get_level4(the_level);
    
}


void get_level1(int **the_level)
{
    int i,j;
    
    // basic level info
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_MAZE_OFFSET] = SCREEN_CENTER_X-TILE_SIZE;    // x pos of top left corner of maze in pixels to start
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_MAZE_OFFSET] = SCREEN_CENTER_Y-TILE_SIZE;    // y coordinate of top left corner of maze in pixels to start
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_CHAR_OFFSET] = 0;                              // x coordinate of starting position of character within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_CHAR_OFFSET] = 0;                              // y coordinate of starting position of character within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_GOAL_OFFSET] = 1;                              // x coordinate of goal position within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_GOAL_OFFSET] = 1;                              // y coordinate of goal position within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_HALF_XWIN_OFFSET] = 3;                              // half window size in x direction
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_HALF_YWIN_OFFSET] = 3;                              // half window size in y direction
    
    // maze tile info
    // first row
    the_level[DESC_MAZE_OFFSET+0][0] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+0][1] = TOP_RIGHT_CORNER;
    for(i = 2; i < MAX_MAZE_SIZE_X; i++ ) the_level[DESC_MAZE_OFFSET+0][i] = NO_MAZE_TILE;
    
    // second row
    the_level[DESC_MAZE_OFFSET+1][0] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+1][1] = BOTTOM_RIGHT_CORNER;
    for(i = 2; i < MAX_MAZE_SIZE_X; i++ ) the_level[DESC_MAZE_OFFSET+1][i] = NO_MAZE_TILE;
    
    // rest of the rows
    for(i = DESC_MAZE_OFFSET+2; i <= MAX_MAZE_SIZE_Y; i++ ) {
        for( j = 0; j < MAX_MAZE_SIZE_X; j++ )
            the_level[i][j] = NO_MAZE_TILE;
    }
            
    // obj tile info
    // set everything to empty first 
    for(i = DESC_OBJ_OFFSET; i < DESC_OBJ_OFFSET+MAX_MAZE_SIZE_Y; i++ ) {
        for( j = 0; j < MAX_MAZE_SIZE_X; j++ )
            the_level[i][j] = NO_OBJ_TILE;
    }
    
}


void get_level2(int **the_level)
{
    int i,j;
    
    // basic level info
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_MAZE_OFFSET] = SCREEN_CENTER_X-3*TILE_SIZE;    // x pos of top left corner of maze in pixels to start
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_MAZE_OFFSET] = SCREEN_CENTER_Y-TILE_SIZE;    // y coordinate of top left corner of maze in pixels to start
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_CHAR_OFFSET] = 2;                              // x coordinate of starting position of character within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_CHAR_OFFSET] = 0;                              // y coordinate of starting position of character within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_GOAL_OFFSET] = 0;                              // x coordinate of goal position within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_GOAL_OFFSET] = 0;                              // y coordinate of goal position within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_HALF_XWIN_OFFSET] = 6;                              // half window size in x direction
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_HALF_YWIN_OFFSET] = 2;                              // half window size in y direction
    
    // maze tile info
    // first row
    the_level[DESC_MAZE_OFFSET+0][0] = LEFT_DEADEND;
    the_level[DESC_MAZE_OFFSET+0][1] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][2] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][3] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][4] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][5] = RIGHT_DEADEND;
    for(i = 6; i < MAX_MAZE_SIZE_X; i++ ) the_level[DESC_MAZE_OFFSET+0][i] = NO_MAZE_TILE;
  
    // rest of the rows
    for(i = DESC_MAZE_OFFSET+1; i <= MAX_MAZE_SIZE_Y; i++ ) {
        for( j = 0; j < MAX_MAZE_SIZE_X; j++ )
            the_level[i][j] = NO_MAZE_TILE;
    }
            
    // obj tile info
    // set everything to empty first 
    for(i = DESC_OBJ_OFFSET; i < DESC_OBJ_OFFSET+MAX_MAZE_SIZE_Y; i++ ) {
        for( j = 0; j < MAX_MAZE_SIZE_X; j++ )
            the_level[i][j] = NO_OBJ_TILE;
    }
    
    // then set where the objects are
    the_level[DESC_OBJ_OFFSET+0][0] = RIGHT_WALL_GREEN;
    the_level[DESC_OBJ_OFFSET+0][1] = KEY_BLUE;
    the_level[DESC_OBJ_OFFSET+0][2] = LEFT_WALL_RED;
    the_level[DESC_OBJ_OFFSET+0][3] = KEY_RED;
    the_level[DESC_OBJ_OFFSET+0][4] = RIGHT_WALL_BLUE;
    the_level[DESC_OBJ_OFFSET+0][5] = KEY_GREEN;
    
}


void get_level3(int **the_level)
{
    int i,j;
    
    // basic level info
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_MAZE_OFFSET] = SCREEN_CENTER_X-2*TILE_SIZE;    // x pos of top left corner of maze in pixels to start
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_MAZE_OFFSET] = SCREEN_CENTER_Y-2*TILE_SIZE;    // y coordinate of top left corner of maze in pixels to start
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_CHAR_OFFSET] = 2;                              // x coordinate of starting position of character within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_CHAR_OFFSET] = 2;                              // y coordinate of starting position of character within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_GOAL_OFFSET] = 4;                              // x coordinate of goal position within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_GOAL_OFFSET] = 2;                              // y coordinate of goal position within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_HALF_XWIN_OFFSET] = 2;                              // half window size in x direction
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_HALF_YWIN_OFFSET] = 2;                              // half window size in y direction
    
    // maze tile info
    // first row
    the_level[DESC_MAZE_OFFSET+0][0] = TOP_DEADEND;
    the_level[DESC_MAZE_OFFSET+0][1] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+0][2] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][3] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][4] = TOP_RIGHT_CORNER;
    for(i = 5; i < MAX_MAZE_SIZE_X; i++ ) the_level[DESC_MAZE_OFFSET+0][i] = NO_MAZE_TILE;
    
    // second row
    the_level[DESC_MAZE_OFFSET+1][0] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+1][1] = LEFT_WALL;
    the_level[DESC_MAZE_OFFSET+1][2] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+1][3] = TOP_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+1][4] = VERTICAL_HALL;
    for(i = 5; i < MAX_MAZE_SIZE_X; i++ ) the_level[DESC_MAZE_OFFSET+1][i] = NO_MAZE_TILE;
    
     // third row
    the_level[DESC_MAZE_OFFSET+2][0] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+2][1] = BOTTOM_DEADEND;
    the_level[DESC_MAZE_OFFSET+2][2] = TOP_DEADEND;
    the_level[DESC_MAZE_OFFSET+2][3] = RIGHT_WALL;
    the_level[DESC_MAZE_OFFSET+2][4] = BOTTOM_DEADEND;
    for(i = 5; i < MAX_MAZE_SIZE_X; i++ ) the_level[DESC_MAZE_OFFSET+2][i] = NO_MAZE_TILE;
    
    // fourth row;
    the_level[DESC_MAZE_OFFSET+3][0] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+3][1] = BOTTOM_WALL;
    the_level[DESC_MAZE_OFFSET+3][2] = RIGHT_WALL;
    the_level[DESC_MAZE_OFFSET+3][3] = NO_MAZE_TILE;
    the_level[DESC_MAZE_OFFSET+3][4] = RIGHT_WALL;
    for(i = 5; i < MAX_MAZE_SIZE_X; i++ ) the_level[DESC_MAZE_OFFSET+3][i] = NO_MAZE_TILE;

    // fifth row
    the_level[DESC_MAZE_OFFSET+4][0] = LEFT_DEADEND;
    the_level[DESC_MAZE_OFFSET+4][1] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+4][2] = BOTTOM_WALL;
    the_level[DESC_MAZE_OFFSET+4][3] = BOTTOM_WALL;
    the_level[DESC_MAZE_OFFSET+4][4] = BOTTOM_DEADEND;
    for(i = 5; i < MAX_MAZE_SIZE_X; i++ ) the_level[DESC_MAZE_OFFSET+4][i] = NO_MAZE_TILE;
    
    // rest of the rows
    for(i = DESC_MAZE_OFFSET+5; i <= MAX_MAZE_SIZE_Y; i++ ) {
        for( j = 0; j < MAX_MAZE_SIZE_X; j++ )
            the_level[i][j] = NO_MAZE_TILE;
    }
            
    // obj tile info
    // set everything to empty first 
    for(i = DESC_OBJ_OFFSET; i < DESC_OBJ_OFFSET+MAX_MAZE_SIZE_Y; i++ ) {
        for( j = 0; j < MAX_MAZE_SIZE_X; j++ )
            the_level[i][j] = NO_OBJ_TILE;
    }
    
    // then set where the objects are
    the_level[DESC_OBJ_OFFSET+0][0] = KEY_RED;
    the_level[DESC_OBJ_OFFSET+0][4] = BOTTOM_WALL_BLUE;
    the_level[DESC_OBJ_OFFSET+1][1] = BOTTOM_WALL_GREEN;
    the_level[DESC_OBJ_OFFSET+2][1] = KEY_BLUE;
    the_level[DESC_OBJ_OFFSET+3][4] = BOTTOM_WALL_RED;
    the_level[DESC_OBJ_OFFSET+4][4] = KEY_GREEN;
    
}




void get_level4(int **the_level)
{
    int i,j;
    
    // basic level info
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_MAZE_OFFSET] = SCREEN_CENTER_X-8*TILE_SIZE;    // x pos of top left corner of maze in pixels to start
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_MAZE_OFFSET] = SCREEN_CENTER_Y-8*TILE_SIZE;    // y coordinate of top left corner of maze in pixels to start
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_CHAR_OFFSET] = 7;                              // x coordinate of starting position of character within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_CHAR_OFFSET] = 7;                              // y coordinate of starting position of character within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_GOAL_OFFSET] = 0;                              // x coordinate of goal position within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_GOAL_OFFSET] = 15;                              // y coordinate of goal position within the maze
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_HALF_XWIN_OFFSET] = 4;                              // half window size in x direction
    the_level[DESC_BASIC_OFFSET][DESC_BASIC_HALF_YWIN_OFFSET] = 4;                              // half window size in y direction
    
    // maze tile info
    // first row
    the_level[DESC_MAZE_OFFSET+0][0] = TOP_DEADEND;
    the_level[DESC_MAZE_OFFSET+0][1] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+0][2] = TOP_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+0][3] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+0][4] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][5] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][6] = TOP_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+0][7] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+0][8] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][9] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][10] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+0][11] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][12] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][13] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+0][14] = RIGHT_DEADEND;
    the_level[DESC_MAZE_OFFSET+0][15] = TOP_DEADEND;
    
    // second row
    the_level[DESC_MAZE_OFFSET+1][0] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+1][1] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+1][2] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+1][3] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+1][4] = LEFT_DEADEND;
    the_level[DESC_MAZE_OFFSET+1][5] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+1][6] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+1][7] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+1][8] = LEFT_DEADEND;
    the_level[DESC_MAZE_OFFSET+1][9] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+1][10] = RIGHT_WALL;
    the_level[DESC_MAZE_OFFSET+1][11] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+1][12] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+1][13] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+1][14] = RIGHT_DEADEND;
    the_level[DESC_MAZE_OFFSET+1][15] = VERTICAL_HALL;
    
    // third row
    the_level[DESC_MAZE_OFFSET+2][0] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+2][1] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+2][2] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+2][3] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+2][4] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+2][5] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+2][6] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+2][7] = NO_MAZE_TILE;
    the_level[DESC_MAZE_OFFSET+2][8] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+2][9] = TOP_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+2][10] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+2][11] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+2][12] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+2][13] = TOP_DEADEND;
    the_level[DESC_MAZE_OFFSET+2][14] = TOP_DEADEND;
    the_level[DESC_MAZE_OFFSET+2][15] = VERTICAL_HALL;
    
    // fourth row
    the_level[DESC_MAZE_OFFSET+3][0] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+3][1] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+3][2] = LEFT_WALL;
    the_level[DESC_MAZE_OFFSET+3][3] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+3][4] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+3][5] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+3][6] = TOP_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+3][7] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+3][8] = TOP_DEADEND;
    the_level[DESC_MAZE_OFFSET+3][9] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+3][10] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+3][11] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+3][12] = LEFT_WALL;
    the_level[DESC_MAZE_OFFSET+3][13] = BOTTOM_WALL;
    the_level[DESC_MAZE_OFFSET+3][14] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+3][15] = VERTICAL_HALL;
    
    // fifth row
    the_level[DESC_MAZE_OFFSET+4][0] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+4][1] = BOTTOM_DEADEND;
    the_level[DESC_MAZE_OFFSET+4][2] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+4][3] = LEFT_DEADEND;
    the_level[DESC_MAZE_OFFSET+4][4] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+4][5] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+4][6] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+4][7] = BOTTOM_DEADEND;
    the_level[DESC_MAZE_OFFSET+4][8] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+4][9] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+4][10] = RIGHT_WALL;
    the_level[DESC_MAZE_OFFSET+4][11] = BOTTOM_DEADEND;
    the_level[DESC_MAZE_OFFSET+4][12] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+4][13] = LEFT_DEADEND;
    the_level[DESC_MAZE_OFFSET+4][14] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+4][15] = BOTTOM_RIGHT_CORNER;
    
    // sixth row
    the_level[DESC_MAZE_OFFSET+5][0] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+5][1] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+5][2] = BOTTOM_WALL;
    the_level[DESC_MAZE_OFFSET+5][3] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+5][4] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+5][5] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+5][6] = BOTTOM_WALL;
    the_level[DESC_MAZE_OFFSET+5][7] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+5][8] = NO_MAZE_TILE;
    the_level[DESC_MAZE_OFFSET+5][9] = TOP_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+5][10] = LEFT_WALL;
    the_level[DESC_MAZE_OFFSET+5][11] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+5][12] = NO_MAZE_TILE;
    the_level[DESC_MAZE_OFFSET+5][13] = RIGHT_DEADEND;
    the_level[DESC_MAZE_OFFSET+5][14] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+5][15] = TOP_DEADEND;
    
    // seventh row
    the_level[DESC_MAZE_OFFSET+6][0] = LEFT_DEADEND;
    the_level[DESC_MAZE_OFFSET+6][1] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+6][2] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+6][3] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+6][4] = RIGHT_DEADEND;
    the_level[DESC_MAZE_OFFSET+6][5] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+6][6] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+6][7] = RIGHT_DEADEND;
    the_level[DESC_MAZE_OFFSET+6][8] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+6][9] = BOTTOM_DEADEND;
    the_level[DESC_MAZE_OFFSET+6][10] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+6][11] = TOP_DEADEND;
    the_level[DESC_MAZE_OFFSET+6][12] = LEFT_WALL;
    the_level[DESC_MAZE_OFFSET+6][13] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+6][14] = RIGHT_WALL;
    the_level[DESC_MAZE_OFFSET+6][15] = VERTICAL_HALL;
    
    // eighth row
    the_level[DESC_MAZE_OFFSET+7][0] = TOP_DEADEND;
    the_level[DESC_MAZE_OFFSET+7][1] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+7][2] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+7][3] = TOP_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+7][4] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+7][5] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+7][6] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+7][7] = LEFT_DEADEND;
    the_level[DESC_MAZE_OFFSET+7][8] = NO_MAZE_TILE;
    the_level[DESC_MAZE_OFFSET+7][9] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+7][10] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+7][11] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+7][12] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+7][13] = TOP_DEADEND;
    the_level[DESC_MAZE_OFFSET+7][14] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+7][15] = RIGHT_WALL;
    
    // ninth row
    the_level[DESC_MAZE_OFFSET+8][0] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+8][1] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+8][2] = LEFT_WALL;
    the_level[DESC_MAZE_OFFSET+8][3] = RIGHT_WALL;
    the_level[DESC_MAZE_OFFSET+8][4] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+8][5] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+8][6] = NO_MAZE_TILE;
    the_level[DESC_MAZE_OFFSET+8][7] = TOP_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+8][8] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+8][9] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+8][10] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+8][11] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+8][12] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+8][13] = LEFT_WALL;
    the_level[DESC_MAZE_OFFSET+8][14] = TOP_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+8][15] = VERTICAL_HALL;
    
    // tenth row
    the_level[DESC_MAZE_OFFSET+9][0] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+9][1] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+9][2] = BOTTOM_DEADEND;
    the_level[DESC_MAZE_OFFSET+9][3] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+9][4] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+9][5] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+9][6] = BOTTOM_DEADEND;
    the_level[DESC_MAZE_OFFSET+9][7] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+9][8] = RIGHT_DEADEND;
    the_level[DESC_MAZE_OFFSET+9][9] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+9][10] = LEFT_DEADEND;
    the_level[DESC_MAZE_OFFSET+9][11] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+9][12] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+9][13] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+9][14] = BOTTOM_DEADEND;
    the_level[DESC_MAZE_OFFSET+9][15] = VERTICAL_HALL;
    
    // eleventh row
    the_level[DESC_MAZE_OFFSET+10][0] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+10][1] = LEFT_WALL;
    the_level[DESC_MAZE_OFFSET+10][2] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+10][3] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+10][4] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+10][5] = LEFT_WALL;
    the_level[DESC_MAZE_OFFSET+10][6] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+10][7] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+10][8] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+10][9] = BOTTOM_WALL;
    the_level[DESC_MAZE_OFFSET+10][10] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+10][11] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+10][12] = TOP_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+10][13] = LEFT_WALL;
    the_level[DESC_MAZE_OFFSET+10][14] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+10][15] = BOTTOM_RIGHT_CORNER;
    
    // twelfth row
    the_level[DESC_MAZE_OFFSET+11][0] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+11][1] = BOTTOM_WALL;
    the_level[DESC_MAZE_OFFSET+11][2] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+11][3] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+11][4] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+11][5] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+11][6] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+11][7] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+11][8] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+11][9] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+11][10] = RIGHT_DEADEND;
    the_level[DESC_MAZE_OFFSET+11][11] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+11][12] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+11][13] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+11][14] = LEFT_DEADEND;
    the_level[DESC_MAZE_OFFSET+11][15] = TOP_RIGHT_CORNER;
    
    // thirteenth row
    the_level[DESC_MAZE_OFFSET+12][0] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+12][1] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+12][2] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+12][3] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+12][4] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+12][5] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+12][6] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+12][7] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+12][8] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+12][9] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+12][10] = TOP_DEADEND;
    the_level[DESC_MAZE_OFFSET+12][11] = BOTTOM_DEADEND;
    the_level[DESC_MAZE_OFFSET+12][12] = BOTTOM_DEADEND;
    the_level[DESC_MAZE_OFFSET+12][13] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+12][14] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+12][15] = RIGHT_WALL;
    
    // fourteenth row
    the_level[DESC_MAZE_OFFSET+13][0] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+13][1] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+13][2] = TOP_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+13][3] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+13][4] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+13][5] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+13][6] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+13][7] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+13][8] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+13][9] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+13][10] = LEFT_WALL;
    the_level[DESC_MAZE_OFFSET+13][11] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+13][12] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+13][13] = TOP_WALL;
    the_level[DESC_MAZE_OFFSET+13][14] = TOP_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+13][15] = BOTTOM_DEADEND;
    
    // fifteenth row
    the_level[DESC_MAZE_OFFSET+14][0] = TOP_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+14][1] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+14][2] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+14][3] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+14][4] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+14][5] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+14][6] = RIGHT_DEADEND;
    the_level[DESC_MAZE_OFFSET+14][7] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+14][8] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+14][9] = BOTTOM_WALL;
    the_level[DESC_MAZE_OFFSET+14][10] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+14][11] = BOTTOM_DEADEND;
    the_level[DESC_MAZE_OFFSET+14][12] = TOP_DEADEND;
    the_level[DESC_MAZE_OFFSET+14][13] = VERTICAL_HALL;
    the_level[DESC_MAZE_OFFSET+14][14] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+14][15] = RIGHT_DEADEND;
    
    // sixeenth row
    the_level[DESC_MAZE_OFFSET+15][0] = BOTTOM_DEADEND;
    the_level[DESC_MAZE_OFFSET+15][1] = LEFT_DEADEND;
    the_level[DESC_MAZE_OFFSET+15][2] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+15][3] = BOTTOM_WALL;
    the_level[DESC_MAZE_OFFSET+15][4] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+15][5] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+15][6] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+15][7] = BOTTOM_WALL;
    the_level[DESC_MAZE_OFFSET+15][8] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+15][9] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+15][10] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+15][11] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+15][12] = BOTTOM_RIGHT_CORNER;
    the_level[DESC_MAZE_OFFSET+15][13] = BOTTOM_LEFT_CORNER;
    the_level[DESC_MAZE_OFFSET+15][14] = HORIZONTAL_HALL;
    the_level[DESC_MAZE_OFFSET+15][15] = RIGHT_DEADEND;
            
    // obj tile info
    // set everything to empty first 
    for(i = DESC_OBJ_OFFSET; i < DESC_OBJ_OFFSET+MAX_MAZE_SIZE_Y; i++ ) {
        for( j = 0; j < MAX_MAZE_SIZE_X; j++ )
            the_level[i][j] = NO_OBJ_TILE;
    }
    
    // then set where the objects are
    the_level[DESC_OBJ_OFFSET+0][0] = KEY_RED;
    the_level[DESC_OBJ_OFFSET+0][15] = KEY_BLUE;
    the_level[DESC_OBJ_OFFSET+6][13] = LEFT_WALL_RED;
    the_level[DESC_OBJ_OFFSET+11][8] = TOP_WALL_BLUE;
    the_level[DESC_OBJ_OFFSET+11][2] = BOTTOM_WALL_GREEN;
    the_level[DESC_OBJ_OFFSET+15][15] = KEY_GREEN;
    
}


// assumes fromx and fromy are valid
int can_move(int **the_level, int fromx, int fromy, int dir) {
    int maze_self, maze_up, maze_down, maze_left, maze_right;
    int obj_self, obj_up, obj_down, obj_left, obj_right;
    
    maze_self = the_level[DESC_MAZE_OFFSET+fromy][fromx];
    obj_self = the_level[DESC_OBJ_OFFSET+fromy][fromx];
    
    switch( dir ) {
        case UP:
            // check maze tile ok
            if( fromy == 0 ) maze_up = -1;
            else maze_up = the_level[DESC_MAZE_OFFSET+fromy-1][fromx];
            
            if( maze_up == -1 || contains_wall(maze_self,UP) || contains_wall(maze_up,DOWN) ) return 0;
            
            // check obj tile ok
            if( fromy == 0 ) obj_up = -1;
            else obj_up = the_level[DESC_OBJ_OFFSET+fromy-1][fromx];
            
            if( obj_up == -1 || contains_door(obj_self,UP) || contains_door(obj_up,DOWN) ) return 0;
            
            return 1;
        case DOWN:
            // check maze tile ok
            if( fromy == MAX_MAZE_SIZE_Y-1 ) maze_down = -1;
            else maze_down = the_level[DESC_MAZE_OFFSET+fromy+1][fromx];
    
            if( maze_down == -1 || contains_wall(maze_self,DOWN) || contains_wall(maze_down,UP) ) return 0;
            
            // check obj tile ok
            if( fromy == MAX_MAZE_SIZE_Y-1 ) obj_down = -1;
            else obj_down = the_level[DESC_OBJ_OFFSET+fromy+1][fromx];
    
            if( obj_down == -1 || contains_door(obj_self,DOWN) || contains_door(obj_down,UP) ) return 0;
            
            return 1;
        case LEFT:
            // check maze tiles ok
            if( fromx == 0 ) maze_left = -1;
            else maze_left = the_level[DESC_MAZE_OFFSET+fromy][fromx-1];
    
            if( maze_left == -1 || contains_wall(maze_self,LEFT) || contains_wall(maze_left,RIGHT) ) return 0;
        
            // check obj tiles ok   
            if( fromx == 0 ) obj_left = -1;
            else obj_left = the_level[DESC_OBJ_OFFSET+fromy][fromx-1];
    
            if( obj_left == -1 || contains_door(obj_self,LEFT) || contains_door(obj_left,RIGHT) ) return 0;
           
            return 1;
        case RIGHT:
            // check maze tiles ok
            if( fromx == MAX_MAZE_SIZE_X-1 ) maze_right = -1;
            else maze_right = the_level[DESC_MAZE_OFFSET+fromy][fromx+1];
    
            if( maze_right == -1 || contains_wall(maze_self,RIGHT) || contains_wall(maze_right,LEFT) ) return 0;
    
            // check obj tiles ok
            if( fromx == MAX_MAZE_SIZE_X-1 ) obj_right = -1;
            else obj_right = the_level[DESC_OBJ_OFFSET+fromy][fromx+1];
    
            if( obj_right == -1 || contains_door(obj_self,RIGHT) || contains_door(obj_right,LEFT) ) return 0;
                    
            return 1;
    }
    return 0;
}

int contains_wall(int tile_type, int dir) {
    if( dir == UP )
        return (tile_type == TOP_WALL || tile_type == TOP_LEFT_CORNER || tile_type == TOP_RIGHT_CORNER || tile_type == HORIZONTAL_HALL || tile_type == LEFT_DEADEND || tile_type == TOP_DEADEND || tile_type == RIGHT_DEADEND );          
    else if( dir == DOWN )
        return (tile_type == BOTTOM_WALL || tile_type == BOTTOM_LEFT_CORNER || tile_type == BOTTOM_RIGHT_CORNER || tile_type == HORIZONTAL_HALL || tile_type == LEFT_DEADEND || tile_type == RIGHT_DEADEND || tile_type == BOTTOM_DEADEND );
    else if( dir == LEFT )
        return (tile_type == LEFT_WALL || tile_type == BOTTOM_LEFT_CORNER || tile_type == TOP_LEFT_CORNER || tile_type == VERTICAL_HALL || tile_type == TOP_DEADEND || tile_type == LEFT_DEADEND || tile_type == BOTTOM_DEADEND );
    else if( dir == RIGHT )
        return (tile_type == RIGHT_WALL || tile_type == BOTTOM_RIGHT_CORNER || tile_type == TOP_RIGHT_CORNER || tile_type == VERTICAL_HALL || tile_type == TOP_DEADEND || tile_type == RIGHT_DEADEND || tile_type == BOTTOM_DEADEND );
        
    return 0;
}

int contains_door(int tile_type, int dir) {
    if( dir == UP )
        return ( (tile_type == TOP_WALL_RED && !has_key(KEY1)) || (tile_type == TOP_WALL_GREEN && !has_key(KEY2)) || (tile_type == TOP_WALL_BLUE && !has_key(KEY3)) );          
    else if( dir == DOWN )
        return ( (tile_type == BOTTOM_WALL_RED && !has_key(KEY1)) || (tile_type == BOTTOM_WALL_GREEN && !has_key(KEY2)) || (tile_type == BOTTOM_WALL_BLUE && !has_key(KEY3)) );
    else if( dir == LEFT )
        return ( (tile_type == LEFT_WALL_RED && !has_key(KEY1)) || (tile_type == LEFT_WALL_GREEN && !has_key(KEY2)) || (tile_type == LEFT_WALL_BLUE && !has_key(KEY3)));
    else if( dir == RIGHT )
        return ( (tile_type == RIGHT_WALL_RED && !has_key(KEY1)) || (tile_type == RIGHT_WALL_GREEN && !has_key(KEY2)) || (tile_type == RIGHT_WALL_BLUE && !has_key(KEY3)));
        
    return 0;
}

void toggle_char_tile()
{
    IOWR(VGA_BASE,WHICH_CHAR_TILE_OFFSET,(IORD(VGA_BASE,WHICH_CHAR_TILE_OFFSET)+1)%4);
}

// TODO make it so that the character moves instead of the maze if moving
// the maze would mean we can see outside the boundaries of the maze
void animate(int dir, wiimote_t *wiimote) {
    int yinc, xinc, num;
    
    int maze_x = get_maze_x();
    int maze_y = get_maze_y();
    
    if( dir == UP ) { yinc = SPEED_PIX; xinc = 0; }
    else if( dir == DOWN ) { yinc = -SPEED_PIX; xinc = 0; }
    else if( dir == LEFT ) { yinc = 0; xinc = SPEED_PIX; }
    else if( dir == RIGHT) { yinc = 0; xinc = -SPEED_PIX; }
    

    for( num = 0; num < TILE_SIZE/SPEED_PIX; num++ )
    {
      /* 
       * Was initially going to be used to play a jingle during gameplay.
       * Due to issues discussed in the final report, this wasn't possible
       */
      //make_moving_sound();
      move_maze_to(maze_x+xinc,maze_y+yinc);
      if( num%4 == 0 ) toggle_char_tile();
      
      // Used to help clear input queue of accelerometer data.
      animate_pause(wiimote);
      
      maze_x += xinc;
      maze_y += yinc;
    } 

}

// returns 1 if x,y is the goal loc, 0 otherwise
int is_at_goal_loc(int **the_level, int x, int y) {
    return (the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_GOAL_OFFSET] == x && the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_GOAL_OFFSET] == y); 
}

// returns which key (KEY1, KEY2, KEY3) is at x,y if not already collected
// or -1 if no key is there (including if it has been collected)
int ontop_of_which_key(int **the_level, int x, int y) {
    if( the_level[DESC_OBJ_OFFSET+y][x] == KEY_RED && !has_key(KEY1) ) return KEY1;
    else if( the_level[DESC_OBJ_OFFSET+y][x] == KEY_GREEN && !has_key(KEY2) ) return KEY2;
    else if( the_level[DESC_OBJ_OFFSET+y][x] == KEY_BLUE && !has_key(KEY3) ) return KEY3;
    return -1;
}

int get_char_start_x(int **the_level) {
    return the_level[DESC_BASIC_OFFSET][DESC_BASIC_XPOS_CHAR_OFFSET];
}

int get_char_start_y(int **the_level) {
    return the_level[DESC_BASIC_OFFSET][DESC_BASIC_YPOS_CHAR_OFFSET];
}

// TODO make this get the wiimote buttons that were pressed
int get_next_move(wiimote_t *wiimote) {
  
  int move=0;
  
  while(1){
    if(wiimote_is_open(wiimote)){
      if(wiimote_update(wiimote)<0){
	wiimote_disconnect(wiimote);
	break;
      }
      
      // Checks for A button.  Turns on accelerometer.
      if (wiimote->keys.a && wiimote->mode.acc == 0) {
	printf("Accelerometer enabled.\n");
	wiimote->mode.acc = 1;
      }

      /*
       * These statements check for certain tilt thresholds.  If they
       * are registered, we move in the respective direction.
       */
      if(wiimote->tilt.x < -50.0) {
	printf("Moving Left, got x acceleration , -50 \n");
	move = LEFT;
	wiimote_update(wiimote);
		break;
      }
            
      if(wiimote->tilt.x > 50.0 ) {
	printf("Moving Right, got x acceleration,  > 50 \n");
	move = RIGHT;
	wiimote_update(wiimote);
	break;
      }
            
      if(wiimote->tilt.y > 30.0 ) {
	printf("Moving Down, got y acceleration, > 30 \n");
	move = DOWN;
	wiimote_update(wiimote);
	break;
      }
      
      if(wiimote->tilt.y < -30.0) {
	printf("Moving Up, got y acceleration, < -30 \n");
	move = UP;
	wiimote_update(wiimote);
	break;
      }
      
      /*
       * These functions provide the same functionality of movement but use
       * the directional pad instead.
       *
       *  if(wiimote->keys.up) {
       *  move = UP;
       *  printf("up was pressed\n");
       *  break;
       *  }
       *  
       *  if(wiimote->keys.down) {
       *  move = DOWN;
       *  printf("down was pressed\n");
       *  break;
       *  }
       *  
       *  if(wiimote->keys.left){
       *  move = LEFT;
       *  printf("left was pressed\n");
       *  break;
       *  }
       *  
       *  if(wiimote->keys.right) {
       *  move = RIGHT;
       *  printf("right was pressed\n");
       *  break;
       *  }
      */

      // Checks if Home key was pressed. Disconnects if it was.
      if(wiimote->keys.home) {
	printf("Home key pressed.  WiiMote disconnected. \n");
	wiimote_disconnect(wiimote);
	exit(0);
      }
    }
  }
  return move;	
}

/*
 * Plays a tone three times when goal is reached
 */
void make_got_to_goal_sound(wiimote_t *wiimote) {
	int i=0;
	for(i=0; i<3; i++) {
		make_got_key_sound(wiimote);
	}
}

/*
 * Checks if the Wiimote is still connected and plays a 20 byte sample
 * if it is. Calls wiimote_update to synchronize
 */
void make_got_key_sound(wiimote_t *wiimote) {
  if (wiimote_is_open(wiimote)) {
    printf("hit key sound \n");
    wiimote_speaker_freq(wiimote, WIIMOTE_FREQ_44800HZ);
    wiimote_speaker_volume(wiimote, 60);
    wiimote_speaker_play(wiimote, sample, 20);
    wiimote_update(wiimote);
  }    
}

/*
 * Was going to be used for jingle sounds.
 * Limitations discussed in final report
 */
void make_moving_sound() {
    
}

/*
 * Checks if the WiiMote is still connected and plays a single
 * 20 byte sample. Calls wiimote_update to synchronize
 */
void make_hit_wall_sound(wiimote_t *wiimote)
{
  if (wiimote_is_open(wiimote)) {
    printf("hit wall sound \n");
    wiimote_speaker_freq(wiimote, 90);
    wiimote_speaker_volume(wiimote, 60);
    wiimote_speaker_play(wiimote, sample, 20);
    wiimote_update(wiimote);
  }
}

void pause() {
    int foo;
    for(foo = 0; foo < PAUSE_AMT; foo++);
}

/*
 * This function was used to help synchronize the wiimote without the
 * processing overhead of wiimote_update.  Helped clear the input queue
 * of packets waiting to be processed by just getting the wiimotes state
 * and doing nothing with it
 */
void animate_pause(wiimote_t *wiimote) {
	wiimote_state_t ev ={{0}};

	int foo;
	int animate_pause_amt = 3;

	for(foo = 0; foo < animate_pause_amt; foo++) {
	  wiimote_get_state(wiimote, &ev);
	}
}

